import re
import queue
import threading
import errno

from stomp.const import EOL
from stomp.const import EOF
from stomp.const import MESSAGE
from stomp.const import SEND
from stomp.const import ERROR


content_length_pattern = re.compile('content-length:(\s+)?(?P<content_length>[0-9]+)(\s+)?\n')


class StreamBuffer(object):
    """Takes an incoming byte stream and parses STOMP frames
    from it.
    """
    framerate = property(lambda self: self._framerate)

    def __init__(self, framerate=25, encoding='utf-8', frame_factory=None,
        frame_queue=None):
        """Initialize a new :class:`StreamBuffer` object.

        Args:
            framerate: number of refreshes of the internal frame buffer,
                per second.
        """
        self._encoding = encoding
        self._framerate = (1/framerate) * 1000
        self._frames = frame_queue or queue.Queue()
        self.__buffer = ''
        self.__frame_factory = frame_factory or (lambda x: x)
        self.__lock = threading.Lock()
        self.__data_received_event = threading.Event()
        self.__thread = threading.Thread(target=self.__main_event_loop)
        self.__thread.daemon = True
        self.__thread.start()

    def read(self, blocking=True, timeout=None):
        """Read a STOMP frame from the queue. If `timeout` is ``None``,
        block until a frame is received, else raise :exc:`EnvironmentError`
        after `timeout` milliseconds.
        """
        try:
            frame = self._frames.get(blocking, timeout)
        except queue.Empty:
            raise EnvironmentError(
                errno.EAGAIN, "No frames received from server.")
        return self.__frame_factory(frame)

    def write(self, byte_stream):
        """Write a sequence of bytes to the internal frame queue."""
        if isinstance(byte_stream, bytes):
            byte_stream = byte_stream.decode(encoding=self._encoding)
        with self.__lock:
            self.__buffer += byte_stream
            self.__data_received_event.set()

    def parse_stream(self, buf):
        """Take an incoming byte stream and parse STOMP frames
        from it. If no frame could be parsed (misaligned or
        partial frame received), add it to our internal buffer
        and wait for more data to come.

        Args:
            byte_stream (bytes): a stream of bytes received from
                a STOMP senders

        """
        ptr = buf.find(EOL)
        if ptr == -1 or (EOF not in buf):
            return (None, buf)
        frame_type = buf[:ptr].strip(EOL)

        # From the spec:
        #
        # If a frame body is present, the SEND, MESSAGE and ERROR frames
        # SHOULD include a content-length header to ease frame parsing.
        # If the frame body contains NULL octets, the frame MUST include
        # a content-length header.
        if frame_type not in (MESSAGE, SEND, ERROR):
            raw_frame = buf[:buf.find(EOF) + 1]
        else:
            # We must find the content-length header and parse the
            # length of the message.
            header = buf[:buf.find(EOL*2) + 2]
            sre = content_length_pattern.search(header, re.DOTALL)
            if sre is not None:
                content_length = int(sre.group('content_length'))
                raw_frame = buf[: len(header) + content_length + 1]

                # Ensure that the message length is smaller than or equal
                # to the buffer, that a complete frame is parsed.
                message_length = len(header) + content_length + 1
                if len(buf) < message_length:
                    # Ugly early return.
                    return None, buf
            else:
                # This might go very wrong if the server sends us
                # malformed messages (TODO).
                raw_frame = buf[:buf.find(EOF) + 1]
        buf = buf[len(raw_frame):]

        assert (EOL*2) in raw_frame, \
            "Valid frames contain two newlines following the headers."
        assert raw_frame[-1] == EOF,\
            "STOMP frames are terminated by EOF, got {0}".format(raw_frame)

        # Remove any newlines from the beginning of the buffer. Shouldn't
        # be there, but reality showed that they are. TODO: find out why.
        while buf and (buf[0] == EOL):
            buf = buf[1:]

        return (raw_frame.strip(EOL), buf)

    def __main_event_loop(self):
        # Wait for incoming data and try to parse it.
        while True:
            if not self.wait_for_data(self.framerate):
                continue
            with self.__lock:
                try:
                    self.__main_event()
                finally:
                    self.__data_received_event.clear()

    def __main_event(self):
        raw_frame, self.__buffer = self.parse_stream(self.__buffer)
        if raw_frame is not None:
            frames = []
            while raw_frame is not None:
                self._frames.put(raw_frame)
                raw_frame, self.__buffer = self.parse_stream(self.__buffer)

    def wait_for_data(self, timeout=None):
        """Blocks until data has been received through the `write()`
        method. Return a boolean when data has been received (``True``)
        or on timeout (False).

        Args:
            timeout (int): optional timeout, in milliseconds.

        Returns:
            bool
        """
        if isinstance(timeout, int):
            timeout /= 1000
        return self.__data_received_event.wait(timeout)
